//-*-c++-*------------------------------------------------------------
// Pointer.h
//
// Author: Qi Yang (qiyang@alum.mit.edu)
//
// A template Smart Pointer class that performs reference counting
// garbage collection and supports subclassed pointers.
//
// See Pointer.cpp in the test package for an example of using this
// class.
//
// Thanks: Stefan Tchekanov (stefant@iname.com). The code here copied
// peices from his SmartPtr class.
//
// HISTORY: 12-09-1999  Original version
//
// You may freely use or modify this code provided that above notice
// and notes on your changes are included in all derived versions.
//
//--------------------------------------------------------------------

#if !defined(_POINTER_HEADER_)
#define _POINTER_HEADER_

#if defined(_MSC_VER)
#pragma warning( disable : 4786)
#endif

//--------------------------------------------------------------------

// CRefCount is a helper to Smart Pointer class P<T> and should never
// be called directly by others

class CRefCount
{
//	Constructors and destructor
public:
	CRefCount(void* ptr_) : _ptr(ptr_), _nCount(0) { }
	virtual ~CRefCount() { }

//	Operations
public:

#if defined(WIN32) || defined(_MSC_VER)
	long IncrRefCount() { return ::InterlockedIncrement(&_nCount); }
	long DecrRefCount() { return ::InterlockedDecrement(&_nCount); }
#else
	long IncrRefCount() { return ++_nCount; }
	long DecrRefCount() { return --_nCount; }
#endif

    long GetRefCount() { return _nCount; }
	void* GetPointer() { return _ptr; }

//	Attributes
private:
	void* _ptr;
	long  _nCount;
};

//--------------------------------------------------------------------

template <class T> class P
{
//	Constructors and destructor

public:
	P() : _ref(NULL) { }
	P(const P& ptr_) : _ref(ptr_._ref) { IncrRefCount(); }
	P(const T& obj_) : _ref(new T(obj_)) { IncrRefCount(); }
	P(T* ptr_) : _ref(ptr_ ? new CRefCount(ptr_) : NULL) {
        IncrRefCount();
    }
	virtual	~P() { DecrRefCount(); }

//	Assignments

public:
	P<T>& operator = (const P& ptr_);
	P<T>& operator = (const T& obj_);
	P<T>& operator = (T* ptr_);

//	Operators

public:
	T* operator -> () { return GetPointer(); }
	T& operator * () { return *(GetPointer()); }

//	Casting

	operator T* () { return GetPointer(); }
	operator const T& () { return *(GetPointer()); }

    CRefCount* Reference() { return _ref; }
    P(CRefCount* ref_) : _ref(ref_) { IncrRefCount(); }

//	Comparison

	bool operator == (const P& ptr) { return _ref == ptr._ref; }
	bool operator != (const P& ptr) { return _ref != ptr._ref; }

	bool operator == (T* ptr_) {
        return _ref	? (_ref->GetPointer() == ptr_) : (ptr_ == NULL);
    }
	bool operator != (T* ptr_) {
        return _ref	? (_ref->GetPointer() != ptr_) : (ptr_ != NULL);
    }

//	Attributes

public:

	bool IsNull() { return _ref == NULL; }
	long GetRefCount() { return _ref ? _ref->GetRefCount() : 0; }

//	Implementation

private:
	void IncrRefCount();
	void DecrRefCount();
    T* GetPointer() {
        return _ref ? static_cast<T*>(_ref->GetPointer()) : (T*)NULL;
    }

private:    
    CRefCount* _ref;
};

//---I M P L E M T A T I O N -----------------------------------------

template<class T>
P<T>& P<T>::operator = (const P& ptr_)
{
	if (_ref != ptr_._ref)
    {
		DecrRefCount();
		_ref = ptr_._ref;
		IncrRefCount();
	}
	return *this;
}

template<class T>
P<T>& P<T>::operator = (T* ptr_)
{
	DecrRefCount();
    if (ptr_)
    {
        _ref = new CRefCount(ptr_);
        IncrRefCount();
    }
    else _ref = NULL;
	return *this;
}

template<class T>
P<T>& P<T>::operator = (const T& obj_)
{
	DecrRefCount();
	_ref = new CRefCount(new T(obj_));
	IncrRefCount();
	return *this;
}

template<class T>
void P<T>::IncrRefCount()
{
	if (_ref) _ref->IncrRefCount();
}

template<class T>
void P<T>::DecrRefCount()
{
	if (_ref && _ref->DecrRefCount() <= 0)
    {
        delete GetPointer();
        delete _ref;
        _ref = NULL;
    }
}

#endif // !defined(_POINTER_HEADER_)
